//$Id: dome.cpp 2.18 1995/02/12 04:54:49 PONDERMATIC Released PONDERMATIC $

/*
DOME.CPP - A program & C++ class for the calculation of geodesic dome
		   properties

	 Copyright (C) 1995  Richard J. Bono

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	 Please direct inquiries, comments and modifications to:
	 Richard J. Bono
	 44 Augusta Rd.
	 Brownsville, TX 78521

	 email: 70324.1712@compuserve.com
					-or-
			rjbono@AOL.com

Revision history:

$Log: dome.cpp $
'Revision 2.18  1995/02/12  04:54:49  PONDERMATIC
'Release Version of Dome.
'Added command line-support
'Added support for PRN & POV-Ray output files
'Rearranged many functions and parameter passing
'
'Revision 2.0  1995/01/01  22:54:59  PONDERMATI
'First public release. Supports only primary symmetry triangle calculations,
'class I (Alternate) breakdowns, & DXF & ASCII file outputs.
'

Acknowledgements & References:
The main reference used in the creation of this code was "Geodesic Math & How
to Use It" by Hugh Kenner, 1976, University of California Press.
ISBN 0-520-02924-0; Library of Congress Catalog Card Number: 74-27292. Many
thanks to Hugh for putting this data in an accessible format.

Also, many thanks to:
	J. F. Nystrom
	My wife Sonia
	My daughter Kathy
	Chris Fearnley
	Kirby Urner
		&
	R. Buckminster Fuller for changing the way I view Universe.

Compliation:
This program was complied and tested using Borland C++ 4.5 using the large
memory model on a Gateway 2000 4DX2-50V.

The release executables were complied using Borland C++ 3.1. The program will
now run on any IBM PC class machine with or withou a math coprocessor.
*/

#include<iostream.h>
#include<iomanip.h>
#include<fstream.h>
#include<stdlib.h>
#include<conio.h>
#include<math.h>
#include<ctype.h>
#include<string.h>

static char rcsid[]="$Id: dome.cpp 2.18 1995/02/12 04:54:49 PONDERMATIC Released PONDERMATIC $";

const double RAD_TO_DEG = 57.2957795130824;
const double DEG_TO_RAD = .0174532925199433;

#define BYTE unsigned char
#define WORD unsigned int

//Structure containing input parameters
struct parameters{
	long freq;
	long classt;
	long polyt;
	long filet;
	char filename[80];
	int verbose_flag;
} cmd_parm;

//Function prototypes
void logo_display(void);
void get_cmd(struct parameters *command, int param_count, char *param[]);
void get_parameters(struct parameters *command);
void main(int argc, char *argv[]);

//--------------geodesic points class
class Geodesic	{
private:
	//Various indices, and temporary variables
	WORD i,
	j, k, index;
	double sphi, stheta, ephi, etheta;

public:
	long frequency, vertex, faces, edges;
	long classtype, polytype;

	//structure defines geodesic coordinate points
	struct coordinates{
		//vertex labels: 0,0 equals zenith
		int A;
		int B;
		//vertex coordinates: 0,0,0 equals zenith
		long xprime;
		long yprime;
		long zprime;

		//vertex spherical coordinates
		double phi;
		double theta;
		//Vertex cartesian coordinates
		double X;
		double Y;
		double Z;
	};

	//structure defines chord data
	struct chords{
		//start point
		int sA;
		int sB;
		double sphi;
		double stheta;

		//end point
		int eA;
		int eB;
		double ephi;
		double etheta;
	};

	//portability note: huge keyword was needed in order to allocate memory for
	//arrays from the far heap in a MSDOS far memory model.
	coordinates huge *pntcrd;  //pointer to coordinate array
	chords huge *edgepts;      //pointer to chord array

	Geodesic(long, long, long);	//constructor
	~Geodesic();		   			//destructor
	void topology(void);	   		//topological abundance function
	void init_crd(void);	   		//Set up coordinate values
	void spherical(void);	   	//calculate spherical coordinates
	double chord_length(double, double, double, double); //chord length function
	void chord_factor(void);   	//calculates chord factors
	void save_dxf(char *);     	//save chord data in DXF format
	void save_ascii(char *);   	//save all coordinate data in ASCII
	void save_POV(char *);			//save data in POV format
	void save_PRN(char *);			//Save raw data in ASCII
	void display_data(void);		//Display data during program execution
	double clean_float(double);	//function cleans up triag zeros
	double rotate_phi(double, double, double, double);	//phi rotation function
	double rotate_theta(double, double, double, double); //theta rotation function
};

//--------------geodesic constructor
Geodesic::Geodesic(long frq, long clss, long ply)
{

	//The constructor handles all calculation when an instance is called.
	//The input parameters are the frequency of subdivision, the class
	//type, and the polygon type. This is all that is needed to
	//specify a geodesic dome! (Note: I have not included radius as size is
	//special case and scaling factors can be applied. The program assumes
	//unit radius domes).

	//initialize class variables
	frequency = frq;
	classtype = clss;
	polytype = ply;
	vertex = 0;
	faces = 0;
	edges = 0;

	//start calculations
	topology();			//calculate symmetry triangle topological abundance
	init_crd();			//initialize vertex coordinates and data structures
	spherical(); 		//calculate spherical coordinates
	chord_factor();  	//determine chord factors
}
//--------------geodesic destructor
Geodesic::~Geodesic() {}
//-------------------calculate chord factors
void Geodesic::chord_factor(void)
{
	//Function cycles through each vertex and determines the
	//chord connections. This code is not very elegant but I think
	//it provides more readablity. Future incarnations may modify this to
	//reduce source code size.

	index = 1;

	for(i=1; i<vertex; i++){
		if(pntcrd[i].A == 0 && pntcrd[i].B == 0){
			//point is the zenith vertex; add 10 & 11
			//line #1 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

			//line #2 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

		}
		else if(pntcrd[i].A == frequency && pntcrd[i].B == 0){
			//point is a left vertex; add 01
			//line #1 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;
		}
		else if(pntcrd[i].A == pntcrd[i].B){
			//point is a right edge; add 10 & 11
			//line #1 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

			//line #2 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;
		}
		else if(pntcrd[i].A == frequency){
			//point is a bottom vertex; add 01
			//line #1 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A  && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;
		}
		else if(pntcrd[i].B == 0){
			//point is a right edge; add 01, 10 & 11
			//line #1 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

			//line #2 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

			//line #3 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;
		}
		else {
			//point is an interior vertex; add 01, 10 & 11
			//line #1 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

			//line #2 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

			//line #3 start point definition
			edgepts[index].sA = pntcrd[i].A;
			edgepts[index].sB = pntcrd[i].B;
			edgepts[index].sphi = pntcrd[i].phi;
			edgepts[index].stheta = pntcrd[i].theta;

			for(j=i; j<=vertex; j++){
				if(pntcrd[j].A == pntcrd[i].A + 1 && pntcrd[j].B == pntcrd[i].B + 1)
					break;
			}
			//end point
			edgepts[index].eA = pntcrd[j].A;
			edgepts[index].eB = pntcrd[j].B;
			edgepts[index].ephi = pntcrd[j].phi;
			edgepts[index].etheta = pntcrd[j].theta;

			index++;

		}
	}

}
//--------------------clean floating point numbers
double Geodesic::clean_float(double fp_number)
{
		//function cleans floating point results close to but not quite zero
		if((fp_number < 0.0) & (fp_number > -.0000000000001))
			fp_number = fabs(fp_number);

		return fp_number;
}
//-------------------save ASCII PRN format
void Geodesic::save_PRN(char *filename)
{
	double sX, sY, sZ, eX, eY, eZ;

	//save chord data for symmetry triangle to PRN file
	//filename should include .PRN extension on MS-DOS systems
	ofstream PRN(filename);

	//Set field widths
	PRN << setiosflags(ios::fixed) << setw(8) << setprecision(6);

	//output PRN data...start with spheres at vertexia points
	for(i=1; i<=vertex; i++){
		//convert spherical to cartesian
		sX = clean_float(cos(pntcrd[i].phi * DEG_TO_RAD) * sin(pntcrd[i].theta * DEG_TO_RAD));
		sY = clean_float(sin(pntcrd[i].phi * DEG_TO_RAD) * sin(pntcrd[i].theta * DEG_TO_RAD));
		sZ = clean_float(cos(pntcrd[i].theta * DEG_TO_RAD));
		//Save data
		PRN << sX << ", " << sY << ", " << sZ << '\n';
	}
	PRN << '\n';

	//Now do the chords
	for(i=1; i<=edges; i++){
		//convert spherical to cartesian
		sX = clean_float(cos(edgepts[i].sphi * DEG_TO_RAD) * sin(edgepts[i].stheta * DEG_TO_RAD));
		sY = clean_float(sin(edgepts[i].sphi * DEG_TO_RAD) * sin(edgepts[i].stheta * DEG_TO_RAD));
		sZ = clean_float(cos(edgepts[i].stheta * DEG_TO_RAD));
		eX = clean_float(cos(edgepts[i].ephi * DEG_TO_RAD) * sin(edgepts[i].etheta * DEG_TO_RAD));
		eY = clean_float(sin(edgepts[i].ephi * DEG_TO_RAD) * sin(edgepts[i].etheta * DEG_TO_RAD));
		eZ = clean_float(cos(edgepts[i].etheta * DEG_TO_RAD));
		//save data
		PRN << sX << ", " << sY << ", " << sZ << ", "
			 << eX << ", " << eY << ", " << eZ << '\n';
	}

	PRN.close();
}
//-------------------save POV-Ray file
void Geodesic::save_POV(char *filename)
{
	double sX, sY, sZ, eX, eY, eZ;

	//save chord data for symmetry triangle to POV file
	//filename should include .POV extension on MS-DOS systems

	ofstream POV(filename);
	//Set field widths
	POV << setiosflags(ios::fixed) << setw(8) << setprecision(6);

	//set up file header
	POV << "//POV-Ray script" << '\n';
	POV << "#include" << '"' << "colors.inc" << '"' << '\n' << '\n';
	POV << "#declare Cam_factor = 6" << '\n';
	POV << "#declare Camera_X = 1 * Cam_factor" << '\n';
	POV << "#declare Camera_Y = 0.5 * Cam_factor" << '\n';
	POV << "#declare Camera_Z = -0.3 * Cam_factor" << '\n';

	POV << "camera { location  <Camera_X, Camera_Y, Camera_Z>" << '\n';
	POV << "		up        <0, 1.0,  0>    right     <-1.33, 0,  0>" << '\n';
	POV << "		direction <0, 0,  3>      look_at   <0, 0, 0> }" << '\n' << '\n';

	POV << "light_source { <Camera_X - 2, Camera_Y + 5 , Camera_Z + 5> color White }" << '\n';
	POV << "light_source { <Camera_X - 2, Camera_Y + 5 , Camera_Z - 3> color White }" << '\n';

	POV << '\n' << "// Background:" << '\n';
	POV << "background {color Gray70}" << '\n' << '\n';

	POV << "// POV script for DOME: " << '\n' << '\n';

	POV << "union {" << '\n';
	//output POV data...start with spheres at vertexia points
	for(i=1; i<=vertex; i++){
		//convert spherical to cartesian
		sX = clean_float(cos(pntcrd[i].phi * DEG_TO_RAD) * sin(pntcrd[i].theta * DEG_TO_RAD));
		sY = clean_float(sin(pntcrd[i].phi * DEG_TO_RAD) * sin(pntcrd[i].theta * DEG_TO_RAD));
		sZ = clean_float(cos(pntcrd[i].theta * DEG_TO_RAD));
		//Save data
		POV << "sphere{<" << sX << ", " << sY << ", " << sZ;
		POV << ">,0.04 pigment {Blue} no_shadow}" << '\n';
	}
		POV << '\n';

	//Now do the chords
	for(i=1; i<=edges; i++){
		//convert spherical to cartesian
		sX = clean_float(cos(edgepts[i].sphi * DEG_TO_RAD) * sin(edgepts[i].stheta * DEG_TO_RAD));
		sY = clean_float(sin(edgepts[i].sphi * DEG_TO_RAD) * sin(edgepts[i].stheta * DEG_TO_RAD));
		sZ = clean_float(cos(edgepts[i].stheta * DEG_TO_RAD));
		eX = clean_float(cos(edgepts[i].ephi * DEG_TO_RAD) * sin(edgepts[i].etheta * DEG_TO_RAD));
		eY = clean_float(sin(edgepts[i].ephi * DEG_TO_RAD) * sin(edgepts[i].etheta * DEG_TO_RAD));
		eZ = clean_float(cos(edgepts[i].etheta * DEG_TO_RAD));
		//save data
		POV << "cylinder{<" << sX << ", " << sY << ", " << sZ << ">, ";
		POV << "<" << eX << ", " << eY << ", " << eZ << ">,0.03 pigment {Red} no_shadow}" << '\n';
	}
	POV << "rotate <0, 70, 0> }" << '\n';

	POV.close();

}
//-------------------save DXF file
void Geodesic::save_dxf(char *filename)
{
	double sX, sY, sZ, eX, eY, eZ;

	//save chord data for symmetry triangle to DXF file
	//filename should include .DXF extension on MS-DOS systems

	ofstream DXF(filename);
	//output DXF data
	DXF << "0" << '\n';
	DXF << "SECTION" << '\n';
	DXF << "2" << '\n';
	DXF << "ENTITIES" << '\n';
	//Set field widths
	DXF << setiosflags(ios::fixed) << setw(8) << setprecision(6);
	for(i=1; i<=edges; i++){
		//convert spherical to cartesian
		sX = clean_float(cos(edgepts[i].sphi * DEG_TO_RAD) * sin(edgepts[i].stheta * DEG_TO_RAD));
		sY = clean_float(sin(edgepts[i].sphi * DEG_TO_RAD) * sin(edgepts[i].stheta * DEG_TO_RAD));
		sZ = clean_float(cos(edgepts[i].stheta * DEG_TO_RAD));
		eX = clean_float(cos(edgepts[i].ephi * DEG_TO_RAD) * sin(edgepts[i].etheta * DEG_TO_RAD));
		eY = clean_float(sin(edgepts[i].ephi * DEG_TO_RAD) * sin(edgepts[i].etheta * DEG_TO_RAD));
		eZ = clean_float(cos(edgepts[i].etheta * DEG_TO_RAD));
		//save data
		DXF << "0" << '\n';
		DXF << "LINE" << '\n';
		DXF << "8" << '\n';
		DXF << "0" << '\n';
		DXF << "10" << '\n';
		DXF << sX << '\n';
		DXF << "20" << '\n';
		DXF << sY << '\n';
		DXF << "30" << '\n';
		DXF << sZ << '\n';
		DXF << "11" << '\n';
		DXF << eX << '\n';
		DXF << "21" << '\n';
		DXF << eY << '\n';
		DXF << "31" << '\n';
		DXF << eZ << '\n';
	}
	DXF << "0" << '\n';
	DXF << "ENDSEC" << '\n';
	DXF << "0" << '\n';
	DXF << "EOF" << '\n';

	DXF.close();

}
//-------------------save data to ASCII file
void Geodesic::save_ascii(char *filename)
{
	//function to save data to ASCII data file
	//WARNING: CAN produce very large listings!
	//save chord data for symmetry triangle to DAT file
	//filename should include .DAT extension on MS-DOS systems

	double sphi, stheta, ephi, etheta;

	ofstream ASCII(filename);
	//output ASCII report data
	ASCII << "Geodesic Dome Data" << '\n';
	ASCII << "----------------Dome Parameters-------------------------" << '\n';
	ASCII << "Polyhedron Type: ";
	if(polytype == 1)
		ASCII << "Icosahedron" << '\n';
	else if(polytype == 2)
		ASCII << "Octahedron" << '\n';
	else
		ASCII << "Tetrahedron" << '\n';
	ASCII << "Class ";
	if(classtype == 1)
		ASCII << "I" << '\n';
	else
		ASCII << "II" << '\n';
	ASCII << "Frequency: " << frequency << '\n';
	ASCII << "----------------Symmetry Triangle Data------------------" << '\n';
	ASCII << "Vertexia: " << vertex << '\n';
	ASCII << "Edges: " << edges << '\n';
	ASCII << "Faces: " << faces << '\n';
	ASCII << "-----------Symmetry Triangle Spherical Coordinates------" << '\n';
	ASCII << "A" << '\t' << "B" << '\t' << "phi" << '\t' << "theta" << '\n';
	for(i=1; i<=vertex; i++){
		ASCII << setiosflags(ios::right) << setw(3);
		ASCII << pntcrd[i].A << '\t';
		ASCII << pntcrd[i].B << '\t';
		ASCII << setw(11) << setprecision(8) << setiosflags(ios::fixed);
		ASCII << pntcrd[i].phi << '\t';
		ASCII << pntcrd[i].theta << '\n' << '\n';
	}
	ASCII << "----------Symmetry Triangle Chord Data------------------" << '\n';
	ASCII << "AB start/AB end" << '\t' << "Chord Length" << '\n';
	for(i=1; i<=edges; i++){
		ASCII << setiosflags(ios::right) << setw(3);
		ASCII << edgepts[i].sA << "," << edgepts[i].sB << "/";
		ASCII << edgepts[i].eA << "," << edgepts[i].eB << '\t';
		ASCII << setw(12) << setprecision(8) << setiosflags(ios::fixed);
		sphi = edgepts[i].sphi;
		stheta = edgepts[i].stheta;
		ephi = edgepts[i].ephi;
		etheta = edgepts[i].etheta;
		ASCII << chord_length(sphi, stheta, ephi, etheta) << '\n';
	}
	ASCII << "End Dome Data" << '\n';

	ASCII.close();
}
//-------------------calculate chord lengths
double Geodesic::chord_length(double sphi, double stheta, double ephi, double etheta)
{
	//Function returns distance between two points given their
	//spherical coordinates

	double result;

	result = pow((2-2 * (cos(stheta * DEG_TO_RAD) * cos(etheta * DEG_TO_RAD) +
		cos((sphi-ephi) *DEG_TO_RAD) * sin(stheta * DEG_TO_RAD) *
			sin(etheta * DEG_TO_RAD))),0.5);

	return result;

}
//------------------phi rotation function
double Geodesic::rotate_phi(double phi, double phi1, double theta1, double theta2)
{
		//phi half of the rotation formula given in Kenner's "Geodesic Math &
		//How to Use it" Kenner credits this formula to Professor George Owen
		double result;

		result = asin((sin(theta1 * DEG_TO_RAD)* sin((phi-phi1) * DEG_TO_RAD) /
						sin(theta2 * DEG_TO_RAD))) * RAD_TO_DEG;

		return result;

}
//------------------theta rotation function
double Geodesic::rotate_theta(double phi, double phi1, double theta, double theta1)
{
		//theta half of the rotation formula given in Kenner's "Geodesic Math &
		//How to Use it" Kenner credits this formula to Professor George Owen
		double result;

		result = acos(cos(theta1 * DEG_TO_RAD) * cos(theta * DEG_TO_RAD) +
						sin(theta1 * DEG_TO_RAD) * sin(theta * DEG_TO_RAD) *
							cos((phi-phi1) * DEG_TO_RAD)) * RAD_TO_DEG;

		return result;

}
//--------------calculate topological abundance of symmetry triangle
void Geodesic::topology(void)
{

	//This class function calculates the topological abundance
	//of the symmetry triangle.

	//calculate the number of vertexia & faces in symmetry triangle
	//This is based on the class type.
	if(classtype == 1){
		for(i=1; i<=frequency; i++){
			if(i == 1)
				vertex = 3;
			else
				vertex += (i + 1);
		}
		faces = frequency * frequency;
	}
	else if(classtype == 2){
		for(i=1; i<=frequency/2; i++){
			if(i == 1)
				vertex = 3;
			else
				vertex += (i + 1);
		}
		faces = (frequency/2) * (frequency/2);
	 }

	//calculate number of edges
	edges = (vertex + faces) - 1;

}
//-------------initialize coordinate array by setting vertex points
void Geodesic::init_crd(void)
{
	//this class function first allocates memory for the coordinate
	//and chord arrays. The vertex labels and vertex coordinates are then
	//calculated. See chapter 12 of Kenner's "Geodesic Math & How to Use
	//it" for more specifics on the labeling & coordinates

	//create array object for vertexia
	pntcrd = new huge coordinates[vertex + 1];
	if (pntcrd == NULL){
		logo_display();
		cout << "Insufficient memory for coordinate array --- Execution Terminating" << '\n';
		exit (0);
	}

	//create array object for edges
	edgepts = new huge chords[edges + 1];
	if (edgepts == NULL){
		logo_display();
		cout << "Insufficient memory for chord array --- Execution Terminating" << '\n';
		exit (0);
	}

	//Initialize vertex labels and coordinates
	index = 1;
	for(i=0; i<=frequency; i++){
		for(j=0; j<=i; j++){
			pntcrd[index].A = i;
			pntcrd[index].B = j;
			pntcrd[index].xprime = j;
			pntcrd[index].yprime = i - j;
			pntcrd[index].zprime = frequency - i;
			index++;
		}
	}
}
//-------------calculate spherical coordinates for given polygon type & class
void Geodesic::spherical(void)
{
	//calculate spherical coordinates of geodesic vertexia
	//given coordinates values
	//care must be taken to avoid generating a trig error

	for(i=1; i<=vertex; i++){
		//set-up x,y,z coordinates for spherical calulations
		//See page 75 of Kenner's "Geodesic Math & How to use it"
		//for formulae used below
		if(polytype == 2){
				//Octahedron
				pntcrd[i].X = pntcrd[i].xprime;
				pntcrd[i].Y = pntcrd[i].yprime;
				pntcrd[i].Z = pntcrd[i].zprime;
		}else if(polytype == 1){
				//Icosahedron
				pntcrd[i].X = pntcrd[i].xprime * sin(72.0 * DEG_TO_RAD);
				pntcrd[i].Y = pntcrd[i].yprime + (pntcrd[i].xprime * cos(72.0 * DEG_TO_RAD));
				pntcrd[i].Z = frequency/2.0 + (pntcrd[i].zprime / (2.0 * cos(36.0 * DEG_TO_RAD)));
		}else if(polytype == 3){
				//Tetrahedron
				pntcrd[i].X = pntcrd[i].xprime * pow(3.0, 0.5);
				pntcrd[i].Y = 2 * pntcrd[i].yprime - pntcrd[i].xprime;
				pntcrd[i].Z = (3.0 * pntcrd[i].zprime - pntcrd[i].xprime - pntcrd[i].yprime) / pow(2.0, 0.5);
		}


		//Calculate phi while avoiding trig errors
		if(pntcrd[i].Y == 0 && pntcrd[i].X == 0)
			pntcrd[i].phi = 0.0;
		else if(pntcrd[i].Y == 0)
			pntcrd[i].phi = 90.0;
		else
			pntcrd[i].phi = atan(pntcrd[i].X / pntcrd[i].Y) * RAD_TO_DEG;

		//adjust value to correct quadrant
		if(pntcrd[i].phi < 0.0)
			pntcrd[i].phi += 180.0;

		//now calculate theta...this is class dependent...
		if(pntcrd[i].Z == 0)
			pntcrd[i].theta = 90.0;
		else if(classtype == 1)
			//All class I types use this equation for theta
			pntcrd[i].theta = atan((pow(pow(pntcrd[i].X, 2.0) + pow(pntcrd[i].Y, 2.0), 0.5)) / pntcrd[i].Z) * RAD_TO_DEG;
		else if(polytype == 2 && classtype == 2)
			//theta for class II octahedra
			pntcrd[i].theta = atan((pow((2.0 * (pow(pntcrd[i].X, 2.0) + pow(pntcrd[i].Y, 2.0))), 0.5)) / pntcrd[i].Z) * RAD_TO_DEG;
		else if(polytype == 1 && classtype == 2)
			//theta for class II Icosahedron
			pntcrd[i].theta = atan((pow(pow(pntcrd[i].X, 2.0) + pow(pntcrd[i].Y, 2.0), 0.5)) / (cos(36.0 * DEG_TO_RAD) * pntcrd[i].Z)) * RAD_TO_DEG;
		else if(polytype == 3 && classtype ==2)
			//theta for class II tetrahedron
			pntcrd[i].theta = atan((pow(pow(pntcrd[i].X, 2.0) + pow(pntcrd[i].Y, 2.0), 0.5)) / (2.0 * pntcrd[i].Z)) * RAD_TO_DEG;

		//make sure the right quadrant is used
		if(pntcrd[i].theta < 0.0)
			pntcrd[i].theta += 180.0;
	}
}
//--------------------display data function
void Geodesic::display_data(void)
{
		cout << "Geodesic Dome Data" << '\n';
		cout << "----------------Dome Parameters-------------------------" << '\n';
		cout << "Polyhedron Type: ";
		if(polytype == 1)
			cout << "Icosahedron" << '\n';
		else if(polytype == 2)
			cout << "Octahedron" << '\n';
		else
			cout << "Tetrahedron" << '\n';
		cout << "Class ";
		if(classtype == 1)
			cout << "I" << '\n';
		else
			cout << "II" << '\n';
		cout << "Frequency: " << frequency << '\n';
		cout << "----------------Symmetry Triangle Data------------------" << '\n';
		cout << "Vertexia: " << vertex << '\n';
		cout << "Edges: " << edges << '\n';
		cout << "Faces: " << faces << '\n' << '\n';
		cout << "Press any key to continue...";

		getch();
		logo_display();

		cout << "-----------Symmetry Triangle Spherical Coordinates------" << '\n';
		cout << "A,B" << '\t' << "phi" << '\t' << "theta" << '\n';
		for(i=1; i<=vertex; i++){
			cout << setiosflags(ios::right) << setw(3);
			cout << pntcrd[i].A << "," << pntcrd[i].B << '\t';
			cout << setw(11) << setprecision(8) << setiosflags(ios::fixed);
			cout << pntcrd[i].phi << '\t';
			cout << pntcrd[i].theta << '\n';
		}
		cout << '\n' << "Press any key to continue..." << '\n';
		getch();

		logo_display();

		cout << "----------Symmetry Triangle Chord Data------------------" << '\n';
		cout << "AB start/AB end" << '\t' << '\t' << "Chord Length" << '\n';
		for(i=1; i<=edges; i++){
			cout << setiosflags(ios::right) << setw(3);
			cout << edgepts[i].sA << "," << edgepts[i].sB << "/";
			cout << edgepts[i].eA << "," << edgepts[i].eB << '\t'<< '\t';
			cout << setw(12) << setprecision(8) << setiosflags(ios::fixed);
			sphi = edgepts[i].sphi;
			stheta = edgepts[i].stheta;
			ephi = edgepts[i].ephi;
			etheta = edgepts[i].etheta;
			cout << '\t' << chord_length(sphi, stheta, ephi, etheta) << '\n';
		}
		cout << "Press any key to continue...";
		getch();

		logo_display();
}
//--------------------End of class Geodesic

void logo_display(void)
{
	clrscr();
	cout << "Dome 2.18, Copyright (C) 1995, Richard J. Bono" << '\n';
	cout << "Dome comes with ABSOLUTELY NO WARRANTY. This is free software," << '\n';
	cout << "and you are welcome to redistribute it under certain conditions." << '\n';
	cout << "See GNU General Public License for more details." << '\n' << '\n';
}
void get_parameters(struct parameters *command)
{
	//geodesic parameters
	WORD badfreq;

	logo_display();

	cout << "Select Polygon type:" << '\n';
	cout << "1. Icosahedron" << '\n';
	cout << "2. Octahedron" << '\n';
	cout << "3. Tetrahderon      ";
	cin >> command->polyt;

	do{
		logo_display();

		cout << "Select Class: " << '\n';
		cout << "1. Class I" << '\n';
		cout << "2. Class II    ";
		cin >> command->classt;
		if(command->classt == 2){
			logo_display();
			cout << "Class II currently not supported" << '\n';
			cout << "Press any key to continue...";
			getch();
		}
	}while (command->classt !=1);

	do{
		logo_display();

		cout << "Enter Frequency of Subdivision: ";
		cin >> command->freq;
		//Only even frequencies are allowed on class 2 domes!
		if(fmod(command->freq, 2) != 0 && command->classt == 2)
			badfreq = 0;
		else
			badfreq = 1;
	}while(!badfreq);

	do{
		logo_display();
		cout << "Output file format " << '\n';
		cout << "1.) DXF" << '\n' << "2.) ASCII" << '\n' << "3.) POV-Ray    ";
		cout << '\n' << "4.) PRN     " << '\n' << "5.) None    ";
		cin >> command->filet;
	}while(command->filet < 1 && command->filet > 5);

	logo_display();
	if(command->filet == 1){
		cout << "Enter DXF filename (.DXF): ";
		cin >> command->filename;
	}
	else if(command->filet == 2){
		cout << "Enter DAT filename (.DAT): ";
		cin >> command->filename;
	}
	else if(command->filet == 3){
		cout << "Enter POV filename (.POV): ";
		cin >> command->filename;
	}
	else if(command->filet == 4){
		cout << "Enter PRN filename (.PRN): ";
		cin >> command->filename;
	}
	command->verbose_flag = 1;
	logo_display();

}
void get_cmd(struct parameters *command, int param_count, char *param[])
{
	//Get and parse command line
	char cmd_parm[6];

	//Set defaults
	command->freq = 2;			//frequency = 2
	command->classt = 1;		//Class I
	command->polyt = 1;			//Icosahedron
	command->verbose_flag = 1;  //verbose output
	command->filet = 5;			//No file output

	int t, j, k;

	for(t=1; t<param_count; ++t){
		if(param[t][0] == '-'){
			switch (tolower(param[t][1])){
				case ('i'):
					//set polyhedron to Icosahedron
					command->polyt = 1;
					break;
				case ('o'):
					//set polyhedron to Octahedron
					command->polyt = 2;
					break;
				case ('t'):
					//set polyhedron to Tetrahedron
					command->polyt = 3;
					break;
				case ('h'):
					//Display help and exit. This overrides other parameters
					logo_display();
					cout << "Usage: dome -fnnn -cn -i -o -t [-v] [-h] [filename.xxx]" << '\n';
					cout << "Where: -fnnn is geodesic frequency" << '\n';
					cout << "             default nnn=2" << '\n';
					cout << "       -cn is class type (n=1 or 2)" << '\n';
					cout << "             default n=1" << '\n';
					cout << "       -i sets Icosahedron (default)" << '\n';
					cout << "       -o sets Octahedron" << '\n';
					cout << "       -t sets Tetrahedron" << '\n';
					cout << "       -v disable data display at run-time" << '\n';
					cout << "       -h displays this help screen" << '\n';
					cout << "        filename.xxx is a standard DOS filename" << '\n';
					cout << "        xxx = DXF, DAT, POV or PRN" << '\n';
					exit(0);
					break;
				case ('v'):
					//Disable Parameter Display during execution
					command->verbose_flag = 0;
					break;
				case ('f'):
					//get frequency
					j=2;
					k=0;
					while(param[t][j]){
						cmd_parm[k] = param[t][j];
						j++;
						k++;
					}
					cmd_parm[k] = '\0';
					command->freq = atol(cmd_parm);
					if (command->freq <= 0){
						logo_display();
						cout << "Invalid Frequency --- Execution Terminating" << '\n';
						exit(1);
					}
					break;
				case ('c'):
					//get class type
					j=2;
					k=0;
					while(param[t][j]){
						cmd_parm[k] = param[t][j];
						j++;
						k++;
					}
					cmd_parm[k] = '\0';
					command->classt = atol(cmd_parm);
					if (command->classt < 1 || command->classt > 2){
						logo_display();
						cout << "Invalid Class Type --- Execution Terminating" << '\n';
						exit(1);
					}
					break;
				default:
					logo_display();
					cout << "Invalid command-line --- Execution Terminating" << '\n';
					exit(1);
					break;
			}
		}
		else{
			//Check to see if parameter is a valid filename
			j = 0;
			while(param[t][j]){
				if(param[t][j] != '.')
					j++;
				else{
					//Check for valid extension
					j++;
					k = 0;
					while(param[t][j]){
						cmd_parm[k] = tolower(param[t][j]);
						j++;
						k++;
					}
					cmd_parm[k] = '\0';
					if(!strcmp(cmd_parm,"dxf"))
						command->filet = 1;
					else if(!strcmp(cmd_parm, "dat"))
						command->filet = 2;
					else if(!strcmp(cmd_parm, "pov"))
						command->filet = 3;
					else if(!strcmp(cmd_parm, "prn"))
						command->filet = 4;
					else{
						logo_display();
						cout << "Invalid command-line --- Execution terminating" << '\n';
						exit(1);
					}
					j = 0;
					k = 0;
					while(param[t][j]){
						command->filename[k] = tolower(param[t][j]);
						j++;
						k++;
					}
					command->filename[k] = '\0';
				}
			}
		}
	}
	if(fmod(command->freq, 2) != 0 && command->classt == 2){
		logo_display();
		cout << "Class II requires Even Frequency --- Execution Terminating" << '\n';
		exit(1);
	}
	if(command->classt == 2){
		logo_display();
		cout << "Class II currently not supported --- Execution Terminating" << '\n';
		exit(1);
	}
}

void main(int argc, char *argv[])
{
	//This main routines shows what can be done with this class. It implements
	//a simple program which displays dome parameters and optionally saves the
	//data in either DXF or ASCII file formats.

	//Some things that can be tried involve creating several instances of the
	//geodesic object. Geodesic shells can be created in this way given enough
	//memory.

	//Command line parameters are changing this routine into something more
	//permanent. Remember the class can be incorporated into your own programs.
	//This shell is just one possible implementation.

	//Get dome parameters
	if(argc == 1)
		//set up interactive input
		get_parameters(&cmd_parm);
	else
		//Get command line
		get_cmd(&cmd_parm, argc, argv);

	//class instance
	Geodesic geosys(cmd_parm.freq, cmd_parm.classt, cmd_parm.polyt);

	logo_display();
	if(cmd_parm.verbose_flag)
		geosys.display_data();
	else
		cout << "Calculating Parameters...Please Wait..." ;

	if(cmd_parm.filet == 1){
		//output file in DXF format
		logo_display();
		geosys.save_dxf(cmd_parm.filename);
	}
	else if(cmd_parm.filet == 2){
		//output all data in ASCII report format
		logo_display();
		geosys.save_ascii(cmd_parm.filename);
	}
	else if(cmd_parm.filet ==3){
		//output data in POV-Ray script format
		logo_display();
		geosys.save_POV(cmd_parm.filename);
	}
	else if(cmd_parm.filet ==4){
		//output data in PRN format
		logo_display();
		geosys.save_PRN(cmd_parm.filename);
	}

	cout << "Execution Complete" << '\n';
}
